# file generated by setuptools_scm
# don't change, don't track in version control
__version__ = version = '0.1.dev1+gcc8fd61.d20230330'
__version_tuple__ = version_tuple = (0, 1, 'dev1', 'gcc8fd61.d20230330')
